'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.resolve = exports.constraints = exports.constraintNames = undefined;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _ramda = require('ramda');

var _ramda2 = _interopRequireDefault(_ramda);

var _ramdaFantasy = require('ramda-fantasy');

var _hmParser = require('hm-parser');

var _hmParser2 = _interopRequireDefault(_hmParser);

var _sanctuaryDef = require('sanctuary-def');

var _sanctuaryDef2 = _interopRequireDefault(_sanctuaryDef);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/* We need a recursion, so: */
/* eslint-disable no-use-before-define */

/* Types are by convention starts with a capital leter, so: */
/* eslint-disable new-cap */

/*
From https://www.npmjs.com/package/hindley-milner-parser-js:

HMP.parse('hello :: Foo a => a -> String');
{
  name: 'hello',
  constraints: [
    {typeclass: 'Foo', typevar: 'a'}],
  type:
    {type: 'function', text: '', children: [
      {type: 'typevar', text: 'a', children: []},
      {type: 'typeConstructor', text: 'String', children: []}]};
*/

// type TypeMap = StrMap Type

//-----------------------------------------------------------------------------
//
// Utilities
//
//-----------------------------------------------------------------------------

var lift = _ramda2.default.map;
var lift2 = _ramda2.default.liftN(2);
var uncurry2 = _ramda2.default.uncurryN(2);

// :: String -> String
var stripNamespace = _ramda2.default.compose(_ramda2.default.last, _ramda2.default.split('/'));

// :: Number -> String
var spellNumber = function spellNumber(x) {
  return {
    1: 'one',
    2: 'two',
    3: 'three',
    4: 'four',
    5: 'five',
    6: 'six',
    7: 'seven',
    8: 'eight',
    9: 'nine'
  }[x] || x.toString();
};

//-----------------------------------------------------------------------------
//
// Type classes
//
//-----------------------------------------------------------------------------

// TypeClassMap -> String -> TypeClass
var lookupTypeClass = function lookupTypeClass(tcm) {
  return function (name) {
    var tc = tcm[name];
    if (!tc) {
      var allTypeClasses = _ramda2.default.keys(tcm).join(', ');
      throw new TypeError('Type class ' + name + ' not found. ' + ('Available type classes are: ' + allTypeClasses));
    }

    return tc;
  };
};

// [SignatureConstraint] -> StrMap String
var constraintNames = exports.constraintNames = _ramda2.default.converge(_ramda2.default.zipObj, [_ramda2.default.pluck('typevar'), _ramda2.default.pluck('typeclass')]);

// TypeClassMap -> [SignatureConstraint] -> StrMap [TypeClass]
var constraints = exports.constraints = uncurry2(function (tcm) {
  return _ramda2.default.compose(_ramda2.default.map(_ramda2.default.of), _ramda2.default.map(lookupTypeClass(tcm)), constraintNames);
});

// :: [TypeClass] -> TypeClassMap
var indexTypeClasses = _ramda2.default.indexBy(_ramda2.default.compose(stripNamespace, _ramda2.default.prop('name')));

//-----------------------------------------------------------------------------
//
// Types
//
//-----------------------------------------------------------------------------

// :: { children :: [a] } -> [a]
var children = _ramda2.default.prop('children');

// :: { children :: [a] } -> a
var firstChild = _ramda2.default.compose(_ramda2.default.prop(0), children);

// :: Object -> String -> Boolean
var typeEq = _ramda2.default.propEq('type');

// :: SignatureEntry -> Boolean
var hasChildren = _ramda2.default.compose(_ramda2.default.not, _ramda2.default.isEmpty, _ramda2.default.prop('children'));

// Clones an object but do not flatten its prototype properties
// as R.clone does
// :: a -> a
var cloneObjPreserveType = function cloneObjPreserveType(obj) {
  // eslint-disable-next-line prefer-const
  var result = Object.create(Object.getPrototypeOf(obj));
  var properties = Object.getOwnPropertyNames(obj);
  for (var i = 0; i < properties.length; i += 1) {
    result[properties[i]] = obj[properties[i]];
  }

  return result;
};

// :: [Type] -> Type -> Type
var substituteTypes = function substituteTypes(argTypes) {
  return function (type) {
    return _ramda2.default.compose(_ramda2.default.reduce(function (t, _ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          argKey = _ref2[0],
          argType = _ref2[1];

      // We’re in reducer, so we can cheat and mutate. The mutation is
      // necessary (versus R.assoc) to preserve the type of `t`
      // eslint-disable-next-line no-param-reassign
      t.types[argKey].type = argType;
      return t;
    }, cloneObjPreserveType(type)), _ramda2.default.zip(_ramda2.default.__, argTypes), _ramda2.default.tap(function (keys) {
      var expected = keys.length;
      var actual = argTypes.length;
      if (expected !== actual) {
        throw new TypeError('Type ' + type.name + ' expects ' + spellNumber(expected) + ' ' + ('argument' + (expected === 1 ? '' : 's') + ', ') + ('got ' + spellNumber(argTypes.length)));
      }
    }), _ramda2.default.prop('keys'))(type);
  };
};

// :: [Type] -> Type|Function -> Type
var constructType = uncurry2(function (argTypes) {
  return _ramda2.default.ifElse(_ramda2.default.is(Function), _ramda2.default.apply(_ramda2.default.__, argTypes), substituteTypes(argTypes));
});

// :: SignatureEntry -> Reader TypeMap Type
var lookupType = function lookupType(entry) {
  return (0, _ramdaFantasy.Reader)(function (typeMap) {
    var name = entry.text;
    var t = typeMap[name];
    if (!t) {
      var allTypes = _ramda2.default.keys(typeMap).join(', ');
      throw new TypeError('Type ' + name + ' not found in env. ' + ('Available types are: ' + allTypes));
    }
    return t;
  });
};

// Helper Type to wipe out thunks
var Thunk = _sanctuaryDef2.default.NullaryType('hm-def/Thunk', '', _ramda2.default.F);

// :: SignatureEntry -> Reader TypeMap Type
var convertTypeConstructor = function convertTypeConstructor(entry) {
  return _ramda2.default.ifElse(hasChildren, _ramda2.default.compose(lift2(constructType)(_ramda2.default.__, lookupType(entry)), convertTypes, children), lookupType)(entry);
};

// :: SignatureEntry -> Reader TypeMap Type
var convertList = _ramda2.default.compose(lift(_sanctuaryDef2.default.Array), convertType, firstChild);

// :: SignatureEntry -> Reader TypeMap Type
var convertFunction = _ramda2.default.compose(lift(_sanctuaryDef2.default.Function), convertTypes, children);

// :: SignatureEntry -> Reader TypeMap (Pair String Type)
var convertRecordField = function convertRecordField(entry) {
  return _ramda2.default.compose(lift(function (valueType) {
    return [entry.text, valueType];
  }), convertType, firstChild)(entry);
};

// :: SignatureEntry -> Reader TypeMap Type
var convertRecord = _ramda2.default.compose(lift(_sanctuaryDef2.default.RecordType), lift(_ramda2.default.fromPairs), _ramda2.default.sequence(_ramdaFantasy.Reader.of), _ramda2.default.map(convertRecordField), children);

// :: SignatureEntry -> Type
var convertTypevar = _ramda2.default.memoize(_ramda2.default.compose(_sanctuaryDef2.default.TypeVariable, _ramda2.default.prop('text')));

// :: SignatureEntry -> (Type -> Type)
var unaryTypevar = _ramda2.default.memoize(_ramda2.default.compose(_sanctuaryDef2.default.UnaryTypeVariable, _ramda2.default.prop('text')));

// :: SignatureEntry -> Reader TypeMap Type
var convertConstrainedType = function convertConstrainedType(entry) {
  return _ramda2.default.compose(lift(unaryTypevar(entry)), convertType, firstChild)(entry);
};

// :: SignatureEntry -> Reader TypeMap Type
function convertType(entry) {
  return _ramda2.default.cond([[typeEq('typeConstructor'), convertTypeConstructor], [typeEq('function'), convertFunction], [typeEq('list'), convertList], [typeEq('record'), convertRecord], [typeEq('constrainedType'), convertConstrainedType], [typeEq('typevar'), _ramda2.default.compose(_ramdaFantasy.Reader.of, convertTypevar)], [typeEq('thunk'), _ramda2.default.always(_ramdaFantasy.Reader.of(Thunk))], [_ramda2.default.T, function (e) {
    throw new Error('Don\'t know what to do with signature entry ' + e.type);
  }]])(entry);
}

// :: [SignatureEntry] -> Reader TypeMap [Type]
function convertTypes(entries) {
  return _ramda2.default.compose(lift(_ramda2.default.reject(_ramda2.default.equals(Thunk))), _ramda2.default.sequence(_ramdaFantasy.Reader.of), _ramda2.default.map(convertType))(entries);
}

// Type -> Type
var ensureParametrized = _ramda2.default.when(_ramda2.default.is(Function), function (fn) {
  return _ramda2.default.apply(fn, _ramda2.default.repeat(_sanctuaryDef2.default.Unknown, fn.length));
});

// :: Type -> String
var shortName = _ramda2.default.compose(stripNamespace, _ramda2.default.prop('name'), ensureParametrized);

// :: [Type] -> TypeMap
var indexTypes = _ramda2.default.indexBy(shortName);

//-----------------------------------------------------------------------------
//
// API
//
//-----------------------------------------------------------------------------

// :: [TypeClass] -> [Type] -> String -> {
//      name :: String,
//      constraints :: StrMap TypeClass,
//      types :: [Type]
//    }
var resolve = exports.resolve = _ramda2.default.curry(function (typeClasses, env, signature) {
  var typeMap = indexTypes(env);
  var typeClassMap = indexTypeClasses(typeClasses);
  var sig = _hmParser2.default.parse(signature);
  var entries = sig.type.children;
  return {
    name: sig.name,
    constraints: constraints(typeClassMap, sig.constraints),
    types: convertTypes(entries).run(typeMap)
  };
});